/*
 * tmgroup.js
 *
 * code by Hemiola SUN
 */

var TabMixGroup = 
{
  enabled : true,
  moving  : false,

  lastColor : -1,
  
  init : function () { 
    TabMixGroup._persistAttribute();
  },
  
  /*
   * Persist tab attributes
   */
  _persistAttribute : function() {
    TabMix._Ss.persistTabAttribute("groupID"); 
    TabMix._Ss.persistTabAttribute("oldGroup");
  },
   
  // add tab aTab to group aID
  setID : function (aTab, aID) {
    if ( aID == null ) {
      aTab.removeAttribute("groupID");
      aTab.groupID = null;
    }
    else {
      aTab.setAttribute("groupID", aID);
      aTab.groupID = aID;
    }
    aTab.removeAttribute("oldGroup");
  },
  
  // return all the tabs in group aGroup
  // argument could be either a tab or a group ID
  //
  getTabsInGroup : function (aGroup) {
    let ID = null;
    try {
      if ( typeof(aGroup) == "string" )
        ID = aGroup;
      else if ( aGroup.nodeName == "tab" )
        ID = aGroup.groupID;
    }
    catch (e) {}
    
    if ( ID == null )
      return [];
    
    return Array.filter(gBrowser.tabs, function(tab) {
             return ( tab.groupID == ID );
           });
  },
  
  // Get the color of group aGroup
  // argument could be either a tab or a group ID
  _getColor : function (aGroup) {
    if ( aGroup == null )
      return null;
      
    let ID = null;
    if ( typeof(aGroup) == "string" )
      ID = aGroup;
    else if ( aGroup.nodeName == "tab" )
      ID = aGroup.groupID;
    
    return ID == null ? null : parseInt(ID.charAt(0));
  },

  pickColor : function (aTab) {
    if (aTab.groupID)
      return this._getColor(aTab.groupID);

    if (aTab.hasAttribute("oldGroup"))
      return this._getColor(aTab.getAttribute("oldGroup"));
    
    let color = ( this.lastColor + 1 ) % 6; // six colors should be enough
    let used = [false, false, false, false, false, false];
    let tabs = gBrowser.visibleTabs;
    for ( let i = 0; i < tabs.length; i++ ) {
      let temp = this._getColor(tabs[i]);
      if ( temp != null )
        used[temp] = true;
    }
    
    if ( used[color] ) {
      for ( let i = 0; i < 6; i++ ) {
        if ( !used[i] )
          return i;
      }
      
      let next = this._getColor(TabMix.nextTab(aTab));
      let prev = this._getColor(TabMix.prevTab(aTab));
      while ( color == next || color == prev ) {
        color = (color + 1) % 6;
      }
    }
    return color;
  },
  
  // add aTab into aRef's group
  // if aRef does not belong to any existing group, group aTab and aRef
  //
  addFromTab : function (aTab, aRef, aColor) {
    let ID = aRef.groupID;
    if ( ID == null ) {
      let oldID = aRef.hasAttribute("oldGroup") ? aRef.getAttribute("oldGroup") : null;
      if ( oldID != null &&
           this._getColor(oldID) != this._getColor(TabMix.prevTab(aRef)) &&
           this._getColor(oldID) != this._getColor(TabMix.nextTab(aRef)) )
        ID = oldID;
      else {
        let id = new Date().getTime();
        let color = aColor || this.pickColor(aRef);
        ID = color + "-" + id;
        this.lastColor = color;
      }
      this.setID(aRef, ID);
    }
    this.setID(aTab, ID);
    this.openStack(ID);
  },
  
  // remove aTab from its group
  removeOneTab : function (aTab) {
    let ID = aTab.groupID;
    if ( ID == null )
      return;

    if ( aTab.hasAttribute("hidden") )
      this.openStack(ID);
    
    aTab.removeAttribute("groupID");
    aTab.groupID = null;
    
    let tabs = this.getTabsInGroup(ID);
    if ( tabs.length == 1 ) {
      setTimeout( function () {
        tabs[0].setAttribute("oldGroup", ID);
        tabs[0].removeAttribute("groupID");
        tabs[0].groupID = null;
      }, 200 );
    }
    else if ( tabs.length == 0 )
      aTab.setAttribute("oldGroup", ID);
  },

  // Expand a tab stack
  //   the argument could be a tab or a group  
  //
  openStack : function (aID) {
    let tabs = this.getTabsInGroup(aID);

    for ( let i = 0; i < tabs.length; i++ )
      tabs[i].removeAttribute("hidden");
  },
  
  // Make a group of tab into stack
  //   the argument must be a tab
  //
  makeStack : function (aTab) {
    let ID = aTab.groupID;
    let tabs = this.getTabsInGroup(ID);
    
    for ( let i = 0; i < tabs.length; i++ )
      tabs[i].setAttribute("hidden", ( tabs[i] == aTab ) ? ( i + "," + tabs.length) : "true");
    // this should be modified when "clicking to make stack" is implied
    gBrowser.selectedTab = aTab;
  },

  // when a tab is moved in front of a tab stack, it would be wrongly inserted into the group
  // we should move it again to the correct position
  //
  // It will always be ok if it's at the right side of the group
  //
  tabBeforeStack : function TMG_tabBeforeStack ( aTab, aStack ) {
    if ( aStack.getAttribute("hidden").indexOf(",") >= 0 ) {
      let value = parseInt( aStack.getAttribute("hidden").split(",")[0] );
      if ( value > 0 )
        TabMix.moveTabTo(aTab, aTab._tPos - value);
    }
  },
  
  // moving a tab stack
  moveStack : function TMG_moveStack ( aTab, oldPos ) {
    // we need a check to prevent infinite loop occurs
    if ( this.moving )
      return;
    
    this.moving = true;

    let nextTab = aTab.nextSibling;
    TabMix.moveTabTo( aTab, oldPos );

    let tabs = this.getTabsInGroup(aTab);
    for ( let i = 0; i < tabs.length; i++ )
      this.moveTabBefore( tabs[i], nextTab, true );
      
    this.moving = false;
  },

  // Move aTab before aTarget which is also a tab
  //   aTarget     = null  : move aTab to the end
  //   shouldCheck = true  : the moved tab should pass TabMix.onTabMove() 
  //                 false : just move the tab and do nothing else
  //
  moveTabBefore : function ( aTab, aTarget, shouldCheck ) {
    shouldCheck = ( false || shouldCheck );
    let newPos = ( aTarget == null ) ? gBrowser.tabs.length - 1 : aTarget._tPos;
    if ( aTarget && newPos > aTab._tPos )
      newPos--;
    if ( newPos != aTab._tPos ) {
      if ( shouldCheck )
        gBrowser.moveTabTo( aTab, newPos );
      else
        TabMix.moveTabTo( aTab, newPos );
    }
  },

  _restoreOldTab : function (aID) {
    let tabs = gBrowser.visibleTabs;
    for ( let i = 0; i < tabs.length; i++ ) {
      if ( tabs[i].getAttribute("oldGroup") == aID ) {
        this.setID( tabs[i], aID );
        return true;
      }
    }
    return false;
  },
  
  restore : function (aTab) {
    aTab._restoring = true;
    let ID = aTab.groupID;
    if ( ID == null && aTab.hasAttribute("oldGroup") )
      ID = aTab.getAttribute("oldGroup");

    if (ID) {
      this.setID( aTab, ID );
      let groupTabs = this.getTabsInGroup(ID);

      // no other tabs exist in the same group, look for the old tabs then
      if ( groupTabs.length <= 1 ) {
        if ( !this._restoreOldTab(ID) )
          // no old tabs neither
          this.removeOneTab(aTab);
      }
    }
  },

  // we cannot call moveRestoredTab() in restore(), because Firefox restore function will
  // move the tab again later. so we call moveRestoredTab() when the restored page is finished loading. 
  // see TabMix.onSSTabRestored()
  //
  moveRestoredTab : function (aTab) {
    let ID = aTab.groupID;
    
    // found aTab's "group-mates", move aTab to correct position and then return
    if (ID) {
      this.openStack(ID);
      let tabs = this.getTabsInGroup(ID);
      if ( tabs[0] == aTab ) // first
        this.moveTabBefore( aTab, tabs[1] );
      else if ( tabs[ tabs.length - 1 ] == aTab ) // last
        this.moveTabBefore( aTab, tabs[ tabs.length - 2 ].nextSibling );
    }
    else {
      // aTab is an orphan
      let nextID = TabMix.nextTab(aTab).groupID;
      let prevID = TabMix.prevTab(aTab).groupID;
      
      // aTab is restored inside an existing group, move aTab behind the group
      if ( nextID && nextID == prevID ) {
        let tabs = this.getTabsInGroup(nextID);
        this.moveTabBefore( aTab, tabs[ tabs.length - 1 ].nextSibling );
      }
    }
    aTab._restoring = null;
  }
}

window.addEventListener("load", TabMixGroup.init, false);
